# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Multi Similarity Loss.
    Multi-Similarity Loss with General Pair Weighting for Deep Metric Learning
    https://arxiv.org/abs/1904.06627
"""
from __future__ import annotations

from typing import Any

import tensorflow as tf

from tensorflow_similarity.algebra import build_masks, masked_max, masked_min
from tensorflow_similarity.distances import Distance, distance_canonicalizer
from tensorflow_similarity.types import FloatTensor, IntTensor

from .metric_loss import MetricLoss
from .utils import logsumexp


def multisimilarity_loss(
    query_labels: IntTensor,
    query_embeddings: FloatTensor,
    key_labels: IntTensor,
    key_embeddings: FloatTensor,
    distance: Distance,
    remove_diagonal: bool = True,
    alpha: float = 2.0,
    beta: float = 40,
    epsilon: float = 0.1,
    lmda: float = 0.5,
    center: float = 1.0,
) -> Any:
    """Multi Similarity loss computations.

    Args:
        query_labels: labels associated with the query embed.
        query_embeddings: Embedded query examples.
        key_labels: labels associated with the key embed.
        key_embeddings: Embedded key examples.
        distance: Which distance function to use to compute the pairwise.
        remove_diagonal: Bool. If True, will set diagonal to False in positive
            pair mask
        alpha: The exponential weight for the positive pairs. Increasing alpha
            makes the logsumexp softmax closer to the max positive pair distance,
            while decreasing it makes it closer to max(P) + log(batch_size).
        beta: The exponential weight for the negative pairs. Increasing beta
            makes the logsumexp softmax closer to the max negative pair distance,
            while decreasing it makes the softmax closer to
            max(N) + log(batch_size).
        epsilon: Used to remove easy positive and negative pairs. We only keep
            positives that we greater than the (smallest negative pair - epsilon)
            and we only keep negatives that are less than the
            (largest positive pair + epsilon).
        lmda: Used to weight the distance. Below this distance, negatives are
            up weighted and positives are down weighted. Similarly, above this
            distance negatives are down weighted and positive are up weighted.
        center: This represents the expected distance value and will be used
            to center the values in the pairwise distance matrix. This is used
            when weighting the positive and negative examples, with the hardest
            examples receiving an up weight and the easiest examples receiving a
            down weight. This should 1 for cosine distances which we expect to
            be between [0,2]. The value will depend on the data for L2 and L1
            distances.

    Returns:
        Loss: The loss value for the current batch.
    """
    # [Label]
    # ! Weirdness to be investigated
    # do not remove this code. It is actually needed for specific situation
    # Reshape label tensor to [batch_size, 1] if not already in that format.
    # labels = tf.reshape(labels, (labels.shape[0], 1))
    batch_size = tf.size(query_labels)

    # [distances]
    pairwise_distances = distance(query_embeddings, key_embeddings)

    # [masks]
    positive_mask, negative_mask = build_masks(
        query_labels,
        key_labels,
        batch_size=batch_size,
        remove_diagonal=remove_diagonal,
    )

    # [pair mining using Similarity-P]
    # This is essentially hard mining the negative and positive pairs.

    # Keep all positives > Min(neg_dist - epsilon).
    neg_min, _ = masked_min(pairwise_distances, negative_mask)
    neg_min = tf.math.subtract(neg_min, epsilon)
    pos_sim_p_mask = tf.math.greater(pairwise_distances, neg_min)
    pos_sim_p_mask = tf.math.logical_and(pos_sim_p_mask, positive_mask)

    # Keep all negatives < Max(pos_dist + epsilon).
    pos_max, _ = masked_max(pairwise_distances, positive_mask)
    pos_max = tf.math.add(pos_max, epsilon)
    neg_sim_p_mask = tf.math.less(pairwise_distances, pos_max)
    neg_sim_p_mask = tf.math.logical_and(neg_sim_p_mask, negative_mask)

    # Mark all pairs where we have both valid negative and positive pairs.
    valid_anchors = tf.math.logical_and(
        tf.math.reduce_any(pos_sim_p_mask, axis=1),
        tf.math.reduce_any(neg_sim_p_mask, axis=1),
    )

    # Cast masks as floats to support multiply
    valid_anchors = tf.cast(valid_anchors, dtype=pairwise_distances.dtype)
    pos_sim_p_mask = tf.cast(pos_sim_p_mask, dtype=pairwise_distances.dtype)
    neg_sim_p_mask = tf.cast(neg_sim_p_mask, dtype=pairwise_distances.dtype)

    # [Weight the remaining pairs using Similarity-S and Similarity-N]
    shifted_distances = pairwise_distances + lmda - center
    pos_dists = alpha * shifted_distances
    neg_dists = -1 * beta * shifted_distances

    # [compute loss]

    # Positive pairs with a distance above 0 will be up weighted.
    p_loss = logsumexp(pos_dists, pos_sim_p_mask)
    p_loss = p_loss / alpha

    # Negative pairs with a distance below 0 will be up weighted.
    n_loss = logsumexp(neg_dists, neg_sim_p_mask)
    n_loss = n_loss / beta

    # Remove any anchors that have empty neg or pos pairs.
    # NOTE: reshape is required here because valid_anchors is [m] and
    #       p_loss + n_loss is [m, 1].
    multisim_loss = tf.math.multiply(p_loss + n_loss, tf.reshape(valid_anchors, (-1, 1)))

    return multisim_loss


@tf.keras.utils.register_keras_serializable(package="Similarity")
class MultiSimilarityLoss(MetricLoss):
    """Computes the multi similarity loss in an online fashion.

    `y_true` must be  a 1-D integer `Tensor` of shape (batch_size,).
    It's values represent the classes associated with the examples as
    **integer  values**.

    `y_pred` must be 2-D float `Tensor`  of L2 normalized embedding vectors.
    you can use the layer `tensorflow_similarity.layers.L2Embedding()` as the
    last layer of your model to ensure your model output is properly
    normalized.
    """

    def __init__(
        self,
        distance: Distance | str = "cosine",
        alpha: float = 2.0,
        beta: float = 40.0,
        epsilon: float = 0.1,
        lmda: float = 0.5,
        center: float = 1.0,
        name: str = "MultiSimilarityLoss",
        **kwargs,
    ):
        """Initializes the Multi Similarity Loss.

        Args:
            distance: Which distance function to use to compute the pairwise
                distances between embeddings. Defaults to 'cosine'.

            alpha: The exponential weight for the positive pairs. Increasing
                alpha makes the logsumexp softmax closer to the max positive
                pair distance, while decreasing it makes it closer to
                max(P) + log(batch_size).

            beta: The exponential weight for the negative pairs. Increasing
                beta makes the logsumexp softmax closer to the max negative
                pair distance, while decreasing it makes the softmax closer
                to max(N) + log(batch_size).

            epsilon: Used to remove easy positive and negative pairs. We only
                keep positives that we greater than the (smallest negative
                pair - epsilon) and we only keep negatives that are less than
                the (largest positive pair + epsilon).

            lmda: Used to weight the distance. Below this distance, negatives
                are up weighted and positives are down weighted. Similarly,
                above this distance negatives are down weighted and positive
                are up weighted.

            center: This represents the expected distance value and will be used
                to center the values in the pairwise distance matrix. This is
                used when weighting the positive and negative examples, with the
                hardest examples receiving an up weight and the easiest examples
                receiving a down weight. This should 1 for cosine distances which
                we expect to be between [0,2]. The value will depend on the data
                for L2 and L1 distances.

            name: Loss name. Defaults to 'MultiSimilarityLoss'.
        """

        # distance canonicalization
        distance = distance_canonicalizer(distance)
        self.distance = distance

        super().__init__(
            multisimilarity_loss,
            name=name,
            distance=distance,
            alpha=alpha,
            beta=beta,
            epsilon=epsilon,
            lmda=lmda,
            center=center,
            **kwargs,
        )



import tensorflow as tf
from absl.testing import parameterized
from tensorflow.keras.losses import Reduction
from tensorflow.python.framework import combinations

from tensorflow_similarity import losses


@tf.function
def softnn_util(y_true, x, temperature=1):
    """
    A simple loop based implementation of soft
    nearest neighbor loss to test the code.
    https://arxiv.org/pdf/1902.01889.pdf
    """

    batch_size = tf.shape(y_true)[0]
    loss = 0.0
    for i in tf.range(batch_size, dtype=tf.int32):
        numerator = 0.0
        denominator = 0.0
        for j in tf.range(batch_size, dtype=tf.int32):
            if i == j:
                continue
            if y_true[i] == y_true[j]:
                numerator += tf.math.exp(-1 * tf.math.reduce_sum(tf.math.square(x[i] - x[j])) / temperature)
            denominator += tf.math.exp(-1 * tf.math.reduce_sum(tf.math.square(x[i] - x[j])) / temperature)
        if numerator == 0.0:
            continue
        loss += tf.math.log(numerator / denominator)
    return -loss / tf.cast(batch_size, tf.float32)


@combinations.generate(
    combinations.combine(
        mode=["graph", "eager"],
    )
)
class SoftNNLossTest(tf.test.TestCase, parameterized.TestCase):
    def test_config(self):
        softnn_obj = losses.SoftNearestNeighborLoss(
            reduction=Reduction.SUM,
            name="softnn_loss",
            distance="cosine",
        )
        self.assertEqual(softnn_obj.distance.name, "cosine")
        self.assertEqual(softnn_obj.name, "softnn_loss")
        self.assertEqual(softnn_obj.reduction, Reduction.SUM)

    @parameterized.parameters((0.1), (0.5), (1), (2), (5), (10), (50))
    def test_all_correct(self, temperature):
        num_inputs = 10
        n_classes = 10

        softnn_obj = losses.SoftNearestNeighborLoss(
            reduction=Reduction.SUM_OVER_BATCH_SIZE,
            temperature=temperature,
        )

        # y_true: labels
        y_true = tf.random.uniform((num_inputs, 1), 0, n_classes, dtype=tf.int32)
        # x: embeddings
        y_preds = tf.random.uniform((num_inputs, 20), 0, 1)

        loss = softnn_obj(y_true, y_preds)
        loss_check = softnn_util(
            y_true,
            y_preds,
            temperature,
        )

        loss_diff = loss - loss_check
        self.assertLess(self.evaluate(loss_diff), 1e-3)



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Multiple negatives ranking loss
    Useful in situations where there are only positive examples
    All other example in a batch is considered as negative
"""
from __future__ import annotations

from typing import Any

import tensorflow as tf

from tensorflow_similarity.distances import Distance, distance_canonicalizer
from tensorflow_similarity.types import FloatTensor

from .metric_loss import MetricLoss
from .utils import logsumexp


def multineg_ranking_loss(
    query_emb: FloatTensor,
    key_emb: FloatTensor,
    scale: float,
    distance: Distance,
) -> Any:
    """Computes the multiple negatives ranking loss.

    Args:
        query_emb: Embedded query examples.
        key_emb: Embedded key examples.
        scale: Float multiplier for scaling loss value
        distance: Which distance function to use to compute the pairwise.

    Returns:
        loss: loss value for the current batch.
    """

    pairwise_sim = distance(query_emb, key_emb) * scale
    if distance.name != "inner_product":
        raise ValueError("Distance must be inner_product")

    sii = tf.linalg.diag_part(pairwise_sim)
    sij_mask = tf.ones(pairwise_sim.shape)
    sij = logsumexp(pairwise_sim, sij_mask)
    examples_loss = -sii + sij

    return examples_loss


@tf.keras.utils.register_keras_serializable(package="Similarity")
class MultiNegativesRankLoss(MetricLoss):
    """Multiple negatives ranking loss

    This loss is useful in cases where only positive pairs
    of embeddings are available. For each example in a batch
    it considers all other examples as negative examples and
    computes the loss.
    See https://arxiv.org/abs/1705.00652 for the original paper.
    """

    def __init__(
        self,
        name: str = "MultipleNegativesRankingLoss",
        distance: Distance | str = "inner_product",
        scale: float = 20,
        **kwargs,
    ):
        """Initializes the MultipleNegativesRankingLoss Loss

        Args:
            `distance`: Which distance function to use.
            `scale`: float value for scaling the loss value.
                     Defaults to 20
            `name`: Loss name. Defaults to MultipleNegativesRankingLoss.
        """
        # distance canonicalization
        distance = distance_canonicalizer(distance)
        self.distance = distance
        self.scale = scale

        super().__init__(multineg_ranking_loss, name=name, distance=distance, **kwargs)

    def call(self, query_emb: FloatTensor, key_emb: FloatTensor) -> FloatTensor:
        """Invokes the `LossFunctionWrapper` instance.

        Args:
          query_emb: Query embeddings
          key_emb: Key embeddings

        Returns:
          Loss values per sample.
        """
        loss: FloatTensor = self.fn(query_emb, key_emb, self.scale, **self._fn_kwargs)
        return loss



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from collections.abc import Mapping

import tensorflow as tf

from .precision_at_k import PrecisionAtK


class MapAtK(PrecisionAtK):
    r"""Mean Average precision - mAP@K is computed as.

    $$
    mAP_i@K = \frac{\sum_{j = 1}^{K} {rel_i_j}\times{P_i@j}}{R}
    $$

    Where: K is the number of neighbors in the i_th query result set.
           P is the rolling precision over the i_th query result set.
           R is the cardinality of the target class.
           rel is the relevance mask (indicator function) for the i_th query.
           i represents the i_th query.
           j represents the j_th ranked query result.

    AP@K is biased towards the top ranked results and is a function of the rank
    (K), the relevancy mask (rel), and the number of indexed examples for the
    class (R). The denominator for the i_th query is set to the number of
    indexed examples (R) for the class associated with the i_th query.

    For example, if the index has has 100 embedded examples (R) of class 'a',
    and our query returns 50 results (K) where the top 10 results are all TPs,
    then the AP@50 will be 0.10; however, if instead the bottom 10 ranked
    results are all TPs, then the AP@50 will be much lower (0.012) because we
    apply a penalty for the 40 FPs that come before the relevant query results.

    This metric is useful when we want to ensure that the top ranked results
    are relevant to the query; however, it requires that we pass a mapping from
    the class id to the number of indexed examples for that class.

    Args:
        r: A mapping from class id to the number of examples in the index,
        e.g., r[4] = 10 represents 10 indexed examples from class 4.

        clip_at_r: If True, set precision at K values to 0.0 for all values
        with rank greater than the count defined in the r mapping, e.g., a
        result set of 10 values, for a query label with 7 indexed examples,
        will have the last 3 precision at K values set to 0.0. Use this to
        compute MAP@R. See section 3.2 of https://arxiv.org/pdf/2003.08505.pdf

        name: Name associated with the metric object, e.g., avg_precision@5

        canonical_name: The canonical name associated with metric, e.g.,
        avg_precision@K

        k: The number of nearest neighbors over which the metric is computed.

        distance_threshold: The max distance below which a nearest neighbor is
        considered a valid match.

        average: {'micro'} Determines the type of averaging performed over the
        queries.

        * 'micro': Calculates metrics globally over all queries.

        drop_closest_lookup: If True, remove the closest lookup before computing
        the metrics. This is used when the query set == indexed set.
    """

    def __init__(
        self,
        r: Mapping[int, int] | None = None,
        clip_at_r: bool = False,
        name: str = "map",
        **kwargs,
    ) -> None:
        if kwargs.get("average") == "macro":
            raise ValueError("Mean Average Precision only supports micro averaging.")

        if "canonical_name" not in kwargs:
            kwargs["canonical_name"] = "map@k"

        super().__init__(r=r, clip_at_r=clip_at_r, name=name, **kwargs)

    @property
    def name(self) -> str:
        if self.clip_at_r:
            return f"{self._name}"

        return super().name

    def _get_matches(self, x):
        tp = tf.math.cumsum(x, axis=1)
        p_at_k = tf.math.divide_no_nan(tp, tf.range(1, x.shape[1] + 1, dtype=tp.dtype))
        p_at_k = tf.math.multiply(x, p_at_k)
        return p_at_k



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Any

from tensorflow_similarity.distances import Distance, distance_canonicalizer
from tensorflow_similarity.types import FloatTensor


class Search(ABC):
    def __init__(
        self,
        distance: Distance | str,
        dim: int,
        verbose: int = 0,
        name: str | None = None,
        **kwargs,
    ):
        """Initializes a nearest neigboors search index.

        Args:
            distance: the distance used to compute the distance between
            embeddings.

            dim: the size of the embeddings.

            verbose: be verbose.
        """
        self.distance: Distance = distance_canonicalizer(distance)
        self.dim = dim
        self.verbose = verbose
        self.name = name if name is not None else self.__class__.__name__

    @abstractmethod
    def add(self, embedding: FloatTensor, idx: int, verbose: int = 1, **kwargs):
        """Add a single embedding to the search index.

        Args:
            embedding: The embedding to index as computed by
            the similarity model.

            idx: Embedding id as in the index table.
            Returned with the embedding to allow to lookup
            the data associated with a given embedding.

        """

    @abstractmethod
    def batch_add(self, embeddings: FloatTensor, idxs: Sequence[int], verbose: int = 1, **kwargs):
        """Add a batch of embeddings to the search index.

        Args:
            embeddings: List of embeddings to add to the index.

            idxs (int): Embedding ids as in the index table. Returned with
            the embeddings to allow to lookup the data associated
            with the returned embeddings.

            verbose: Be verbose. Defaults to 1.
        """

    @abstractmethod
    def lookup(self, embedding: FloatTensor, k: int = 5) -> tuple[list[int], list[float]]:
        """Find embedding K nearest neighboors embeddings.

        Args:
            embedding: Query embedding as predicted by the model.
            k: Number of nearest neighboors embedding to lookup. Defaults to 5.
        """

    @abstractmethod
    def batch_lookup(self, embeddings: FloatTensor, k: int = 5) -> tuple[list[list[int]], list[list[float]]]:
        """Find embeddings K nearest neighboors embeddings.

        Args:
            embedding: Batch of query embeddings as predicted by the model.
            k: Number of nearest neighboors embedding to lookup. Defaults to 5.
        """

    @abstractmethod
    def save(self, path: str):
        """Serializes the index data on disk

        Args:
            path: where to store the data
        """

    @abstractmethod
    def load(self, path: str):
        """load index on disk

        Args:
            path: where to store the data
        """

    def get_config(self) -> dict[str, Any]:
        """Contains the search configuration.

        Returns:
            A Python dict containing the configuration of the search obj.
        """
        config = {
            "distance": self.distance.name,
            "dim": self.dim,
            "verbose": self.verbose,
            "name": self.name,
            "canonical_name": self.__class__.__name__,
        }

        return config



import math

import numpy as np
import tensorflow as tf

from tensorflow_similarity.callbacks import EvalCallback
from tensorflow_similarity.classification_metrics import Recall
from tensorflow_similarity.evaluators import MemoryEvaluator
from tensorflow_similarity.models import SimilarityModel


def test_eval_init_defaults():
    queries = tf.constant([[1.0, 2.0], [3.0, 4.0]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1.0, 2.0], [3.0, 4.0]])
    target_labels = tf.constant([1, 2])

    callback = EvalCallback(
        queries=queries,
        query_labels=query_labels,
        targets=targets,
        target_labels=target_labels,
    )

    assert tf.math.reduce_all(callback.queries_known == queries)
    assert tf.math.reduce_all(callback.query_labels_known == query_labels)
    assert tf.math.reduce_all(callback.targets == targets)
    assert tf.math.reduce_all(callback.target_labels == target_labels)
    assert callback.distance == "cosine"
    assert isinstance(callback.evaluator, MemoryEvaluator)
    assert {"binary_accuracy"} == set([m.name for m in callback.classification_metrics])
    assert callback.k == 1
    assert tf.math.reduce_all(callback.distance_thresholds == tf.constant([math.inf]))
    assert callback.matcher == "match_nearest"
    assert callback.tb_writer is None


def test_eval_init(tmp_path):
    queries = tf.constant([[1.0, 2.0], [3.0, 4.0]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1.0, 2.0], [3.0, 4.0]])
    target_labels = tf.constant([1, 2])
    distance = "l2"
    metrics = [Recall()]
    k = 11
    matcher = "majority_vote"
    distance_thresholds = tf.constant([0.1, 0.2, 0.3])

    log_dir = tmp_path / "sec/"
    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    callback = EvalCallback(
        queries=queries,
        query_labels=query_labels,
        targets=targets,
        target_labels=target_labels,
        distance=distance,
        metrics=metrics,
        k=k,
        matcher=matcher,
        distance_thresholds=distance_thresholds,
        tb_logdir=log_dir,
    )

    assert tf.math.reduce_all(callback.queries_known == queries)
    assert tf.math.reduce_all(callback.query_labels_known == query_labels)
    assert tf.math.reduce_all(callback.targets == targets)
    assert tf.math.reduce_all(callback.target_labels == target_labels)
    assert callback.distance == "l2"
    assert isinstance(callback.evaluator, MemoryEvaluator)
    assert {"recall"} == set([m.name for m in callback.classification_metrics])
    assert callback.k == 11
    assert tf.math.reduce_all(callback.distance_thresholds == distance_thresholds)
    assert callback.matcher == "majority_vote"
    assert isinstance(callback.tb_writer, tf.summary.SummaryWriter)


def test_eval_callback(tmp_path):
    queries = tf.constant([[1.0, 2.0], [1.0, 2.0]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1.0, 2.0], [1.0, 2.0]])
    target_labels = tf.constant([1, 2])

    log_dir = tmp_path / "sec/"
    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    callback = EvalCallback(
        queries=queries,
        query_labels=query_labels,
        targets=targets,
        target_labels=target_labels,
        tb_logdir=str(log_dir),
        metrics=["binary_accuracy", "f1score"],
    )

    # manually set model ^^
    tf.keras.backend.clear_session()
    inputs = tf.keras.layers.Input(shape=(2,))
    outputs = tf.keras.layers.Dense(2)(inputs)
    model = SimilarityModel(inputs, outputs)
    model.compile("adam", loss="mse", distance="cosine")
    callback.model = model

    # call the only callback method implemented
    logs = {}
    callback.on_epoch_end(0, logs)

    metric_values = np.array(list(logs.values()))
    assert {"binary_accuracy", "f1score"} == logs.keys()
    np.testing.assert_allclose(np.array([0.5, 0.666667]), metric_values, rtol=1e-5, atol=0)


def test_split_val_init_defaults():
    queries = tf.constant([[1.0, 2.0], [3.0, 4.0]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1.0, 2.0], [3.0, 4.0]])
    target_labels = tf.constant([1, 2])
    known_classes = tf.constant([1])
    q_known = tf.constant([[1.0, 2.0]])
    q_label_known = tf.constant([1], dtype="int32")
    q_unknown = tf.constant([[3.0, 4.0]])
    q_label_unknown = tf.constant([2], dtype="int32")

    callback = EvalCallback(
        queries=queries,
        query_labels=query_labels,
        targets=targets,
        target_labels=target_labels,
        known_classes=known_classes,
    )

    assert tf.math.reduce_all(callback.targets == targets)
    assert tf.math.reduce_all(callback.target_labels == target_labels)
    assert callback.distance == "cosine"
    assert isinstance(callback.evaluator, MemoryEvaluator)
    assert {"binary_accuracy"} == set([m.name for m in callback.classification_metrics])
    assert callback.k == 1
    assert tf.math.reduce_all(callback.distance_thresholds == tf.constant([math.inf]))
    assert callback.matcher == "match_nearest"
    assert callback.tb_writer is None

    assert tf.math.reduce_all(callback.queries_known == q_known)
    assert tf.math.reduce_all(callback.query_labels_known == q_label_known)
    assert tf.math.reduce_all(callback.queries_unknown == q_unknown)
    assert tf.math.reduce_all(callback.query_labels_unknown == q_label_unknown)


def test_split_val_init(tmp_path):
    queries = tf.constant([[1.0, 2.0], [3.0, 4.0]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1.0, 2.0], [3.0, 4.0]])
    target_labels = tf.constant([1, 2])
    known_classes = tf.constant([1])
    distance = "l2"
    metrics = [Recall()]
    k = 11
    matcher = "majority_vote"
    distance_thresholds = tf.constant([0.1, 0.2, 0.3])
    q_known = tf.constant([[1.0, 2.0]])
    q_label_known = tf.constant([1], dtype="int32")
    q_unknown = tf.constant([[3.0, 4.0]])
    q_label_unknown = tf.constant([2], dtype="int32")

    log_dir = tmp_path / "sec/"
    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    callback = EvalCallback(
        queries=queries,
        query_labels=query_labels,
        targets=targets,
        target_labels=target_labels,
        known_classes=known_classes,
        distance=distance,
        metrics=metrics,
        k=k,
        matcher=matcher,
        distance_thresholds=distance_thresholds,
        tb_logdir=log_dir,
    )

    assert tf.math.reduce_all(callback.targets == targets)
    assert tf.math.reduce_all(callback.target_labels == target_labels)
    assert callback.distance == "l2"
    assert isinstance(callback.evaluator, MemoryEvaluator)
    assert {"recall"} == set([m.name for m in callback.classification_metrics])
    assert callback.k == 11
    assert tf.math.reduce_all(callback.distance_thresholds == distance_thresholds)
    assert callback.matcher == "majority_vote"
    assert isinstance(callback.tb_writer, tf.summary.SummaryWriter)

    assert tf.math.reduce_all(callback.queries_known == q_known)
    assert tf.math.reduce_all(callback.query_labels_known == q_label_known)
    assert tf.math.reduce_all(callback.queries_unknown == q_unknown)
    assert tf.math.reduce_all(callback.query_labels_unknown == q_label_unknown)


def test_split_val_loss_callback(tmp_path):
    queries = tf.constant([[1.0, 2.0], [1.0, 2.0]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1.0, 2.0], [1.0, 2.0]])
    target_labels = tf.constant([1, 2])
    known_classes = tf.constant([1])

    log_dir = tmp_path / "sec/"
    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    callback = EvalCallback(
        queries=queries,
        query_labels=query_labels,
        targets=targets,
        target_labels=target_labels,
        known_classes=known_classes,
        tb_logdir=str(log_dir),
    )

    # manually set model ^^
    tf.keras.backend.clear_session()
    inputs = tf.keras.layers.Input(shape=(2,))
    outputs = tf.keras.layers.Dense(2)(inputs)
    model = SimilarityModel(inputs, outputs)
    model.compile("adam", loss="mse", distance="cosine")
    callback.model = model

    # call the only callback method implemented
    callback.on_epoch_end(0, {})


def test_split_val_eval_init(tmp_path):
    queries = tf.constant([[1.0, 2.0], [3.0, 4.0]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1.0, 2.0], [3.0, 4.0]])
    target_labels = tf.constant([1, 2])
    known_classes = tf.constant([1])
    distance = "l2"
    metrics = [Recall()]
    k = 11
    matcher = "majority_vote"
    distance_thresholds = tf.constant([0.1, 0.2, 0.3])
    q_known = tf.constant([[1.0, 2.0]])
    q_label_known = tf.constant([1], dtype="int32")
    q_unknown = tf.constant([[3.0, 4.0]])
    q_label_unknown = tf.constant([2], dtype="int32")

    log_dir = tmp_path / "sec/"
    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    callback = EvalCallback(
        queries=queries,
        query_labels=query_labels,
        targets=targets,
        target_labels=target_labels,
        known_classes=known_classes,
        distance=distance,
        metrics=metrics,
        k=k,
        matcher=matcher,
        distance_thresholds=distance_thresholds,
        tb_logdir=log_dir,
    )

    assert tf.math.reduce_all(callback.targets == targets)
    assert tf.math.reduce_all(callback.target_labels == target_labels)
    assert callback.distance == "l2"
    assert isinstance(callback.evaluator, MemoryEvaluator)
    assert {"recall"} == set([m.name for m in callback.classification_metrics])
    assert callback.k == 11
    assert tf.math.reduce_all(callback.distance_thresholds == distance_thresholds)
    assert callback.matcher == "majority_vote"
    assert isinstance(callback.tb_writer, tf.summary.SummaryWriter)

    assert tf.math.reduce_all(callback.queries_known == q_known)
    assert tf.math.reduce_all(callback.query_labels_known == q_label_known)
    assert tf.math.reduce_all(callback.queries_unknown == q_unknown)
    assert tf.math.reduce_all(callback.query_labels_unknown == q_label_unknown)


def test_split_val_eval_callback(tmp_path):
    queries = tf.constant([[1.0, 2.0], [1.0, 2.0]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1.0, 2.0], [1.0, 2.0]])
    target_labels = tf.constant([1, 2])
    known_classes = tf.constant([1])

    log_dir = tmp_path / "sec/"
    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    callback = EvalCallback(
        queries=queries,
        query_labels=query_labels,
        targets=targets,
        target_labels=target_labels,
        known_classes=known_classes,
        tb_logdir=str(log_dir),
    )

    # manually set model ^^
    tf.keras.backend.clear_session()
    inputs = tf.keras.layers.Input(shape=(2,))
    outputs = tf.keras.layers.Dense(2)(inputs)
    model = SimilarityModel(inputs, outputs)
    model.compile("adam", loss="mse", distance="cosine")
    callback.model = model

    # call the only callback method implemented
    callback.on_epoch_end(0, {})



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"Various utilities functions for improved quality of life."
from __future__ import annotations

import math
from collections.abc import Mapping, MutableMapping, Sequence
from typing import Any

import numpy as np
import tensorflow as tf

from tensorflow_similarity.types import BoolTensor, FloatTensor, IntTensor, Lookup


def is_tensor_or_variable(x):
    "check if a variable is tf.Tensor or tf.Variable"
    return tf.is_tensor(x) or isinstance(x, tf.Variable)


def tf_cap_memory():
    "Avoid TF to hog memory before needing it"
    gpus = tf.config.experimental.list_physical_devices("GPU")

    if gpus:
        for gpu in gpus:
            try:
                tf.config.experimental.set_memory_growth(gpu, True)
            except RuntimeError as e:
                # Memory growth must be set before GPUs have been initialized
                print(e)


def unpack_lookup_labels(lookups: Sequence[Sequence[Lookup]], dtype: str | tf.DType) -> IntTensor:
    # Using list comprehension as it is faster
    all_values = [[n.label for n in lu] for lu in lookups]
    # Lookup sets are not guaranteed to all be the same size. Therefore we load
    # the list of lists as a ragged tensor and convert to an int tensor with a
    # default class label value set to the max value for int.
    base_type = tf.dtypes.as_dtype(dtype).base_dtype
    ragged_labels = tf.ragged.constant(all_values, dtype=base_type)

    if not _same_length_rows(ragged_labels):
        print(
            f"WARNING: {_count_of_small_lookup_sets(ragged_labels)} lookup "
            "sets are shorter than the max lookup set length. Imputing "
            "0x7FFFFFFF for the missing label lookups."
        )

    result: IntTensor = ragged_labels.to_tensor(default_value=0x7FFFFFFF)

    return result


def unpack_lookup_distances(
    lookups: Sequence[Sequence[Lookup]], dtype: str | tf.DType, distance_rounding: int | None = None
) -> FloatTensor:
    # using list comprehension as it is faster
    all_values = [[n.distance for n in lu] for lu in lookups]
    # Lookup sets are not guaranteed to all be the same size. Therefore we load
    # the list of lists as a ragged tensor and convert to an flat32 tensor with a
    # default dist value set to math.inf.
    base_type = tf.dtypes.as_dtype(dtype).base_dtype
    ragged_dists = tf.ragged.constant(all_values, dtype=base_type)

    if distance_rounding is not None:
        multiplier = tf.constant([10.0**distance_rounding])
        ragged_dists = tf.round(ragged_dists * multiplier) / multiplier

    if not _same_length_rows(ragged_dists):
        print(
            f"WARNING: {_count_of_small_lookup_sets(ragged_dists)} lookup "
            "sets are shorter than the max lookup set length. Imputing "
            "math.inf for the missing distance lookups."
        )

    dists: FloatTensor = ragged_dists.to_tensor(default_value=math.inf)

    return dists


def unpack_results(
    results: Mapping[str, np.ndarray],
    epoch: int,
    logs: MutableMapping[str, Any],
    tb_writer: tf.summary.SummaryWriter,
    name_suffix: str | None = "",
) -> list[str]:
    """Updates logs, writes summary, and returns list of strings of
    evaluation metric"""
    mstr = []
    for metric_name, vals in results.items():
        float_val = vals[0] if isinstance(vals, np.ndarray) else vals
        full_metric_name = f"{metric_name}{name_suffix}"
        logs[full_metric_name] = float_val
        mstr.append(f"{full_metric_name}: {float_val:.4f}")
        if tb_writer:
            with tb_writer.as_default():
                tf.summary.scalar(full_metric_name, float_val, step=epoch)
    return mstr


def _same_length_rows(x: tf.RaggedTensor) -> BoolTensor:
    """Check if the rows are all the same length.

    Args:
        x: A RaggedTensor

    Returns:
       True if all rows are the same length.
    """
    dims = tf.expand_dims(x.row_lengths(), axis=-1)
    pairwise_equality = tf.equal(dims, tf.transpose(dims))
    is_same_length: BoolTensor = tf.math.reduce_all(pairwise_equality)
    return is_same_length


def _count_of_small_lookup_sets(x: tf.RaggedTensor) -> IntTensor:
    """The count of lookup sets smaller than x.bounding_shape()[1]

    Args:
        x: A RaggedTensor

    Returns:
        The count of smaller lookup sets.
    """
    rl = x.row_lengths()
    max_rl = x.bounding_shape()[1]
    short_lookup_sets = tf.cast(rl != max_rl, dtype="int32")
    small_lookup_count: IntTensor = tf.math.reduce_sum(short_lookup_sets)
    return small_lookup_count



import tensorflow as tf
from absl.testing import parameterized
from tensorflow.keras.losses import Reduction
from tensorflow.python.framework import combinations

from tensorflow_similarity import losses

from . import utils


@combinations.generate(
    combinations.combine(
        mode=["graph", "eager"],
    )
)
class PNLossTest(tf.test.TestCase, parameterized.TestCase):
    def test_config(self):
        pn_obj = losses.PNLoss(
            reduction=Reduction.SUM,
            name="pn_loss",
            distance="cosine",
        )
        self.assertEqual(pn_obj.distance.name, "cosine")
        self.assertEqual(pn_obj.name, "pn_loss")
        self.assertEqual(pn_obj.reduction, Reduction.SUM)

    @parameterized.named_parameters(
        {
            "testcase_name": "_soft_margin",
            "margin": None,
            "expected_loss": 0.31326169,
        },
        {
            "testcase_name": "_fixed_margin",
            "margin": 1.1,
            "expected_loss": 0.1,
        },
    )
    def test_all_correct_unweighted(self, margin, expected_loss):
        """Test that assumes a perfect embedding.

        Using cosine distance, the prefect embedding will contain one hot encoded
        embeddings with a value of 1.0 in the class position and 0 everywhere else.

        Using Soft Margin we expect `pos_dist - neg_dist = -1.0` yielding
        `ln(1 + e^(-1.0)) = 0.31326169`.

        Using a fixed margin of 1.1 we expect the loss to stop at 0.1.
        """
        pn_obj = losses.PNLoss(
            reduction=Reduction.SUM_OVER_BATCH_SIZE,
            margin=margin,
        )
        y_true, y_preds = utils.generate_perfect_test_batch(batch_size=4)
        loss = pn_obj(y_true, y_preds)
        self.assertAlmostEqual(self.evaluate(loss), expected_loss, 3)

    @parameterized.named_parameters(
        {
            "testcase_name": "_soft_margin",
            "margin": None,
            "expected_loss": 0.7706615,
        },
        {
            "testcase_name": "_fixed_margin",
            "margin": 1.0,
            "expected_loss": 1.125,
        },
    )
    def test_all_mismatch_unweighted(self, margin, expected_loss):
        """Test that assumes none of the classes match.

        Using cosine distance, each embedding is one hot encoded with a value of
        1.0 in the class position and 0 everywhere else.

        Using Soft Margin we expect `pos_dist - neg_dist = 0.0` yielding
        `ln(1 + e^(0.0)) = 0.69314718`. However, the distance between one of the
        positive and negative pairs is 0.0 which is less than the distance of 1.0
        between the anchor and negative example. This means that the loss for this
        pair is `ln(1 + e^(1.0)) = 1.31326169`.

        Using a fixed margin of 1.0 we expect the loss to stop at 1.0. However,
        the distance between one of the positive and negative pairs is 0.0,
        which is less than the distance of 1.0 between the anchor and the negative
        example. This means that the loss for this pair will be `(1.0 - 0.0) + 1.0 = 2.0`.
        """
        pn_obj = losses.PNLoss(
            reduction=Reduction.SUM_OVER_BATCH_SIZE,
            margin=margin,
        )
        y_true, y_preds = utils.generate_bad_test_batch(batch_size=4)
        loss = pn_obj(y_true, y_preds)
        self.assertAlmostEqual(self.evaluate(loss), expected_loss, 3)

    @parameterized.named_parameters(
        {
            "testcase_name": "_soft_margin",
            "margin": None,
            "expected_loss": 0.31326169,
        },
        {
            "testcase_name": "_fixed_margin",
            "margin": 1.1,
            "expected_loss": 0.1,
        },
    )
    def test_no_reduction(self, margin, expected_loss):
        pn_obj = losses.PNLoss(
            reduction=Reduction.NONE,
            margin=margin,
        )
        y_true, y_preds = utils.generate_perfect_test_batch(batch_size=4)
        loss = pn_obj(y_true, y_preds)
        loss = self.evaluate(loss)
        expected_loss = self.evaluate(tf.fill(y_true.shape, expected_loss))
        self.assertArrayNear(loss, expected_loss, 1e-3)

    @parameterized.named_parameters(
        {
            "testcase_name": "_soft_margin",
            "margin": None,
            "expected_loss": 0.31326169,
        },
        {
            "testcase_name": "_fixed_margin",
            "margin": 1.1,
            "expected_loss": 0.1,
        },
    )
    def test_sum_reduction(self, margin, expected_loss):
        pn_obj = losses.PNLoss(
            reduction=Reduction.SUM,
            margin=margin,
        )
        y_true, y_preds = utils.generate_perfect_test_batch(batch_size=4)
        loss = pn_obj(y_true, y_preds)
        expected_loss = y_true.shape[0] * expected_loss
        self.assertAlmostEqual(self.evaluate(loss), expected_loss, 3)

    @parameterized.named_parameters(
        {
            "testcase_name": "_easy",
            "positive_mining_strategy": "easy",
            "negative_mining_strategy": "easy",
            "expected_loss": 0.6666667,
        },
        {
            "testcase_name": "_hard",
            "positive_mining_strategy": "hard",
            "negative_mining_strategy": "hard",
            "expected_loss": 2.0,
        },
    )
    def test_mining(self, positive_mining_strategy, negative_mining_strategy, expected_loss):
        """Test the easy mining strategy

        Easy mining strategy will always pick the minimum distance for the
        positive or negative pair. This means that the pair for positive
        should always be [1.0, 0.0] and for negative should always be [0.0, 1.0].
        However, the distance between two of the positive and negative pairs is 0.0
        which is less than the distance of 1.0 between the anchor and negative example.
        This means that the loss for these pairs is `(1.0 - 0.0) + 1.0 = 2` while the
        loss for the other examples is `(0.0 - 1.0) + 1.0 = 0` and the loss across the
        batch is sum([0.0, 0.0, 2.0, 0.0, 0.0, 2.0]) / 6 = 0.666.

        Hard ming strategy will always pick the maximum distance for the
        positive or negative pair. This means that the pair for positive
        should always be [0.0, 1.0] and for negative should always be [1.0, 0.0].
        This means that the loss should be
        sum([1.0, 1.0, 1.0, 1.0, 1.0, 1.0]) / 6 = 2.0.
        """
        pn_obj = losses.PNLoss(
            reduction=Reduction.SUM_OVER_BATCH_SIZE,
            positive_mining_strategy=positive_mining_strategy,
            negative_mining_strategy=negative_mining_strategy,
            margin=1.0,
        )
        labels = tf.expand_dims(
            tf.constant([0, 0, 0, 1, 1, 1], dtype=tf.int32),
            axis=1,
        )
        embeddings = tf.constant(
            [
                # Class 0
                [1.0, 0.0],
                [1.0, 0.0],
                [0.0, 1.0],
                # Class 1
                [0.0, 1.0],
                [0.0, 1.0],
                [1.0, 0.0],
            ],
            dtype=tf.float32,
        )
        loss = pn_obj(labels, embeddings)
        self.assertAlmostEqual(self.evaluate(loss), expected_loss, 3)

    def test_semi_hard_mining(self):
        """Test the easy positive mining and semi-hard negative mining strategy.

        Easy positive mining strategy will just pick the other positive example,
        i.e., the pos_idxs will be [1,0,3,2]

        Semi-hard negative mining strategy will pick the closest negative example
        that is > the max distance between the positive pairs, if one exists,
        otherwise it will pick the maximal negative example. The neg_idxs will be
        [3,2,1,1] where the second positive example actully has no negatives that
        are > the max distance between the maximal positive.This is why the loss
        for the second positive example is > 1.0.

        The expected pairwise distances are:
            [[0.0,  0.5,  1.0,     0.98],   # Choose 0.98 because it's > 0.5 but < 1.0
             [0.5,  0.0,  0.13,    0.12],   # Choose 0.13 because it's there is no > 0.5
             [1.0,  0.13, 0.0,     0.00005],# Choose 0.13 because it's > 5e-5 but < 1.0
             [0.98, 0.12, 0.00005, 0.0]]    # Choose 0.12 because it's > 5e-5 but < 0.98

        However, two of the positive examples are closer to the negative examples and
        so we get the following negative distances instead of the above: [0.12, 0.13, 0.12, 0.12]
        """
        pn_obj = losses.PNLoss(
            reduction=Reduction.NONE,
            positive_mining_strategy="easy",
            negative_mining_strategy="semi-hard",
            margin=1.0,
        )
        labels = tf.expand_dims(
            tf.constant([0, 0, 1, 1], dtype=tf.int32),
            axis=1,
        )
        embeddings = tf.constant(
            [
                # Class 0
                [1.0, 0.0],
                [0.5, 0.8660254],  # sqrt(3) / 2 or l2 norm of [1.0, 1.0]
                # Class 1
                [0.0, 1.0],
                [0.01010049, 0.999949],
            ],
            dtype=tf.float32,
        )
        loss = pn_obj(labels, embeddings)
        expected_loss = tf.constant(
            [
                [1.371],  # This is the loss betwen the pos and neg instead of between the anchor and neg.
                [1.366],  # This example has no negatives that are > the max distance between the maximal positive.
                [0.871],  # This is the loss betwen the pos and neg instead of between the anchor and neg.
                [0.871],
            ],
            dtype=tf.float32,
        )
        self.assertArrayNear(self.evaluate(loss), self.evaluate(expected_loss), 1e-3)



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""PN Loss foudn in Ivis.
    Szubert, B., Cole, J.E., Monaco, C. et al.
    Structure-preserving visualisation of high dimensional single-cell dataset
    Sci Rep 9, 8914 (2019). https://doi.org/10.1038/s41598-019-45301-0
"""
from __future__ import annotations

from typing import Any

import tensorflow as tf

from tensorflow_similarity.algebra import build_masks
from tensorflow_similarity.distances import Distance, distance_canonicalizer
from tensorflow_similarity.types import FloatTensor, IntTensor

from .metric_loss import MetricLoss
from .utils import compute_loss, negative_distances, positive_distances


def pn_loss(
    query_labels: IntTensor,
    query_embeddings: FloatTensor,
    key_labels: IntTensor,
    key_embeddings: FloatTensor,
    distance: Distance,
    remove_diagonal: bool = True,
    positive_mining_strategy: str = "hard",
    negative_mining_strategy: str = "semi-hard",
    margin: float | None = None,
) -> Any:
    """Positive Negative loss computations.

    Based on the pn loss used in IVIS.

    Args:
        query_labels: labels associated with the query embed.
        query_embeddings: Embedded query examples.
        key_labels: labels associated with the key embed.
        key_embeddings: Embedded key examples.
        distance: The distance function to use to compute the pairwise distances
            between embeddings.
        remove_diagonal: If True, set the diagonal to False in the positive pair
            mask
        positive_mining_strategy: What mining strategy to use to select
            embedding from the same class. {'easy', 'hard'}
        negative_mining_strategy: What mining strategy to use for select the
            embedding from the different class. {'hard', 'semi-hard', 'easy'}
        margin: Defines the target margin between positive and negative pairs,
            e.g., a margin of 1.0 means that the positive and negative
            distances should be 1.0 apart. If None, then a soft margin is used.
            A soft margin can be beneficial to pull together samples from the
            same class as much as possible. See the paper for more details
            https://arxiv.org/pdf/1703.07737.pdf. Defaults to None.

    Returns:
        Loss: The loss value for the current batch.
    """

    # [Label]
    # ! Weirdness to be investigated
    # do not remove this code. It is actually needed for specific situation
    # Reshape label tensor to [batch_size, 1] if not already in that format.
    # labels = tf.reshape(labels, (labels.shape[0], 1))
    batch_size = tf.size(query_labels)

    # [distances]
    pairwise_distances = distance(query_embeddings, key_embeddings)

    # [masks]
    positive_mask, negative_mask = build_masks(
        query_labels,
        key_labels,
        batch_size=batch_size,
        remove_diagonal=remove_diagonal,
    )

    # [Positive distance computation]
    pos_distances, pos_idxs = positive_distances(
        positive_mining_strategy,
        pairwise_distances,
        positive_mask,
    )

    # [Negative distances computation]
    neg_distances, neg_idxs = negative_distances(
        negative_mining_strategy,
        pairwise_distances,
        negative_mask,
        positive_mask,
    )

    # Compute the distance between the pairs of positive and negative examples.
    # Then take the min(pos_neg_dist, anchor_neg_dist) as the neg_distances.
    # This encourages both the anchor and the positives to be far from the
    # negative.
    pn_pairs = tf.stack([pos_idxs, neg_idxs], axis=1)
    pn_distances = tf.gather_nd(pairwise_distances, pn_pairs)
    pn_distances = tf.reshape(pn_distances, [-1, 1])
    neg_distances = tf.math.minimum(pn_distances, neg_distances)

    # [PN loss computation]
    pn_loss = compute_loss(pos_distances, neg_distances, margin)

    return pn_loss


@tf.keras.utils.register_keras_serializable(package="Similarity")
class PNLoss(MetricLoss):
    """Computes the PN loss in an online fashion.

    This loss encourages the positive distances between a pair of embeddings
    with the same labels to be smaller than the minimum negative distances
    between pair of embeddings of different labels. Additionally, both the
    anchor and the positive embeddings are encouraged to be far from the
    negative embeddings. This is accomplished by taking the
    min(pos_neg_dist, anchor_neg_dist) and using that as the negative distance
    in the triplet loss.

    See PN Loss Ivis:
    Szubert, B., Cole, J.E., Monaco, C. et al.
    Structure-preserving visualisation of high dimensional single-cell dataset
    Sci Rep 9, 8914 (2019). https://doi.org/10.1038/s41598-019-45301-0

    `y_true` must be  a 1-D integer `Tensor` of shape (batch_size,).
    It's values represent the classes associated with the examples as
    **integer  values**.

    `y_pred` must be 2-D float `Tensor`  of L2 normalized embedding vectors.
    you can use the layer `tensorflow_similarity.layers.L2Embedding()` as the
    last layer of your model to ensure your model output is properly
    normalized.
    """

    def __init__(
        self,
        distance: Distance | str = "cosine",
        positive_mining_strategy: str = "hard",
        negative_mining_strategy: str = "semi-hard",
        margin: float | None = None,
        name: str = "PNLoss",
        **kwargs,
    ):
        """Initializes the PN Loss.

        Args:
            distance: The distance function to use to compute the pairwise
                distances between embeddings.
            positive_mining_strategy: What mining strategy to use to select
                embedding from the same class. {'easy', 'hard'}
            negative_mining_strategy: What mining strategy to use for select the
                embedding from the different class. {'hard', 'semi-hard', 'easy'}
            margin: Defines the target margin between positive and negative pairs,
                e.g., a margin of 1.0 means that the positive and negative
                distances should be 1.0 apart. If None, then a soft margin is used.
                A soft margin can be beneficial to pull together samples from the
                same class as much as possible. See the paper for more details
                https://arxiv.org/pdf/1703.07737.pdf. Defaults to None.
            name: Loss name. Defaults to "PNLoss".

        Raises:
            ValueError: Invalid positive mining strategy.
            ValueError: Invalid negative mining strategy.
        """

        # distance canonicalization
        distance = distance_canonicalizer(distance)
        self.distance = distance

        # sanity checks
        if positive_mining_strategy not in ["easy", "hard"]:
            raise ValueError("Invalid positive mining strategy.")

        if negative_mining_strategy not in ["easy", "hard", "semi-hard"]:
            raise ValueError("Invalid negative mining strategy.")

        super().__init__(
            pn_loss,
            name=name,
            distance=distance,
            positive_mining_strategy=positive_mining_strategy,
            negative_mining_strategy=negative_mining_strategy,
            margin=margin,
            **kwargs,
        )



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from __future__ import annotations

from collections.abc import Callable

import tensorflow as tf

from tensorflow_similarity.types import Tensor


def random_apply(func: Callable[[Tensor], Tensor], p: float, x: Tensor) -> Tensor:
    """Randomly apply function func to x with probability p."""
    return tf.cond(
        tf.less(
            tf.random.uniform([], minval=0, maxval=1, dtype=tf.float32),
            tf.cast(p, tf.float32),
        ),
        lambda: func(x),
        lambda: x,
    )



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from collections.abc import Mapping

import tensorflow as tf

from tensorflow_similarity.types import BoolTensor, FloatTensor, IntTensor

from .retrieval_metric import RetrievalMetric


class PrecisionAtK(RetrievalMetric):
    r"""Precision@K is computed as.

    $$
    P_i@k = \frac{TP_i}{TP_i+FP_i} = \frac{\sum_{j = 1}^{k} {rel_i_j}}{K}
    $$

    Where: K is the number of neighbors in the i_th query result set.
           rel is the relevance mask (indicator function) for the i_th query.
           i represents the i_th query.
           j represents the j_th ranked query result.

    P@K is unordered and does not take into account the rank of the TP results.

    This metric is useful when we are interested in evaluating the embedding
    within the context of a kNN classifier or as part of a clustering method.

    Args:
        r: An optional mapping from class id to the number of examples in the
        index, e.g., r[4] = 10 represents 10 indexed examples from class 4.
        If None, the value for k will be used instead.

        clip_at_r: If True, set precision at K values to 0.0 for all values
        with rank greater than the count defined in the r mapping, e.g., a
        result set of 10 values, for a query label with 7 indexed examples,
        will have the last 3 precision at K values set to 0.0. Use this to
        compute R Precision. See section 3.2 of https://arxiv.org/pdf/2003.08505.pdf

        name: Name associated with the metric object, e.g., precision@5

        canonical_name: The canonical name associated with metric,
        e.g., precision@K

        k: The number of nearest neighbors over which the metric is computed.

        distance_threshold: The max distance below which a nearest neighbor is
        considered a valid match.

        average: {'micro'} Determines the type of averaging performed over the
        queries.

        * 'micro': Calculates metrics globally over all queries.
        * 'macro': Calculates metrics for each label and takes the unweighted
        mean.

        drop_closest_lookup: If True, remove the closest lookup before computing
        the metrics. This is used when the query set == indexed set.
    """

    def __init__(
        self,
        r: Mapping[int, int] | None = None,
        clip_at_r: bool = False,
        name: str = "precision",
        **kwargs,
    ) -> None:
        if "canonical_name" not in kwargs:
            kwargs["canonical_name"] = "precision@k"

        super().__init__(name=name, **kwargs)
        self.r = r
        self.clip_at_r = clip_at_r

    @property
    def name(self) -> str:
        if self.clip_at_r:
            return f"{self._name}"

        return super().name

    def get_config(self):
        config = {
            "r": self.r,
            "clip_at_r": self.clip_at_r,
        }
        base_config = super().get_config()
        return {**base_config, **config}

    def _generate_class_count_table(self, class_labels, dtype):
        r = self.r
        if r is None:
            r = {int(lbl): self.k for lbl in class_labels}

        if self.drop_closest_lookup:
            r = {k: tf.math.maximum(0, v - 1) for k, v in r.items()}

        return tf.lookup.StaticHashTable(
            tf.lookup.KeyValueTensorInitializer(
                list(r.keys()),
                list(r.values()),
                key_dtype=dtype,
                value_dtype=dtype,
            ),
            default_value=0,
        )

    def _get_matches(self, x):
        return x

    def compute(
        self,
        *,  # keyword only arguments see PEP-570
        query_labels: IntTensor,
        match_mask: BoolTensor,
        **kwargs,
    ) -> FloatTensor:
        """Compute the metric

        Args:
            query_labels: A 1D array of the labels associated with the
            embedding queries.

            match_mask: A 2D mask where a 1 indicates a match between the
            jth query and the kth neighbor and a 0 indicates a mismatch.

            **kwargs: Additional compute args.

        Returns:
            A rank 0 tensor containing the metric.
        """
        self._check_shape(query_labels, match_mask)
        class_labels = tf.unique(tf.reshape(query_labels, (-1)))[0]
        start_k = 0

        if self.drop_closest_lookup:
            start_k = 1

        k_slice = tf.cast(match_mask[:, start_k : self.k], dtype=tf.keras.backend.floatx())
        matches = self._get_matches(k_slice)

        class_count_table = self._generate_class_count_table(class_labels, query_labels.dtype)
        class_counts = class_count_table.lookup(query_labels)

        if self.clip_at_r:
            elems = (matches, tf.expand_dims(class_counts, axis=1))
            # for each row i in matches, reduce_sum the first class_count[i] values and divide by
            # class_count[i]
            per_example = tf.map_fn(
                lambda x: tf.math.divide_no_nan(
                    tf.math.reduce_sum(x[0][: x[1][0]]),
                    tf.cast(x[1], dtype=tf.keras.backend.floatx()),
                ),
                elems,
                fn_output_signature="float",
            )
        else:
            per_example = tf.math.divide_no_nan(
                tf.math.reduce_sum(matches, axis=1),
                tf.cast(class_counts, dtype=tf.keras.backend.floatx()),
            )

        if self.average == "micro":
            avg_res = tf.math.reduce_mean(per_example)
        elif self.average == "macro":
            per_class_metrics = 0
            for label in class_labels:
                idxs = tf.where(query_labels == label)
                c_slice = tf.gather(per_example, indices=idxs)
                per_class_metrics += tf.math.reduce_mean(c_slice)
            avg_res = tf.math.divide_no_nan(per_class_metrics, len(class_labels))
        else:
            raise ValueError(f"{self.average} is not a supported average " "option")
        result: FloatTensor = avg_res
        return result



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contrastive learning specialized losses.
"""
from .barlow import Barlow  # noqa
from .circle_loss import CircleLoss  # noqa
from .metric_loss import MetricLoss  # noqa
from .multinegrank_loss import MultiNegativesRankLoss  # noqa
from .multisim_loss import MultiSimilarityLoss  # noqa
from .pn_loss import PNLoss  # noqa
from .simclr import SimCLRLoss  # noqa
from .simsiam import SimSiamLoss  # noqa
from .softnn_loss import SoftNearestNeighborLoss  # noqa
from .triplet_loss import TripletLoss  # noqa
from .vicreg import VicReg  # noqa
from .xbm_loss import XBM  # noqa
